package goaml.lang.driver;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import goaml.lang.StateChart;
import goaml.lang.entity.Entity;
import goaml.lang.state.AbstractState;
import goaml.lang.state.State;
import goaml.lang.util.Configuration;

public class PlannedUpdate {

	private enum Action {
		ADD, RESET, SUBSTRACT;
	}

	public static final Action ADD = Action.ADD;
	public static final Action RESET = Action.RESET;
	public static final Action SUBSTRACT = Action.SUBSTRACT;

	private final int DEFAULT_NUM_COHORT = -1;
	private final boolean DEFAULT_LOOP = false;
	private final int DEFAULT_MAX_ITERATION = -1;
	private final double DEFAULT_DELAY = 1.0;
	private final double DEFAULT_COEFFICIENT = 1.0;

	private final double START_TIME;
	private final State TARGET;
	private final Action ACTION;
	private final boolean LOOP;
	private final int MAX_ITERATION;
	private final double DELAY;
	private final double COEFFICIENT;

	private double lastValue;
	private int currentIteration;

	private PlannedUpdate(final double startTime, final State target, final Action action, final double value,
			final boolean loop, final int maxIteration, final double delay, final double coefficient) {
		this.START_TIME = startTime;
		this.TARGET = target;
		this.ACTION = action;
		this.LOOP = loop;
		this.MAX_ITERATION = maxIteration + 1;
		this.DELAY = delay;
		this.COEFFICIENT = coefficient;

		this.lastValue = value;
		this.currentIteration = 0;
	}

	// This function parse a CSV file in order to create PlannedUpdate
	public static List<PlannedUpdate> parseCSV(final String path, final StateChart stateChart) {

		List<PlannedUpdate> plannedUpdates = new ArrayList<PlannedUpdate>();

		BufferedReader reader = null;
		try {
			reader = new BufferedReader(new FileReader(path));
			String line = "";
			while ((line = reader.readLine()) != null) {
				String[] values = line.split(String.valueOf(Configuration.CSV_SEPARATOR));
				if (values[0].equals("") || values[0].substring(0, 1).equals("#")) {
					// if the line begins with #, it's a comment, so the line is ignored
					// if the line is empty, it's ignored too
					continue;
				}

				double startTime = Double.valueOf(values[0]);
				List<Entity> entities = stateChart.getEntitiesByName(values[1]);
				String targetName = values[2];
				int numCohort = Integer.valueOf(values[3]);
				Action action = Action.valueOf(values[4]);
				double value = Double.valueOf(values[5]);
				boolean loop = Boolean.valueOf(values[6]);
				int maxIteration = Integer.valueOf(values[7]);
				double delay = Double.valueOf(values[8]);
				double coefficient = Double.valueOf(values[9]);

				// For each state with name equals to "values[2]" (targetName) of each entity with
				// name "values[1]"
				for (Entity entity : entities) {
					List<AbstractState> targets = entity.getStatesByName(targetName);
					for (AbstractState target : targets) {
						if (target instanceof State) {
							// A target must be an instance of State
							if (numCohort == target.getNumCohort()) {
								// Create a PlannedUpdate only if the target has the good numCohort
								plannedUpdates.add(new PlannedUpdate(startTime, (State) target, action, value, loop,
										maxIteration, delay, coefficient));
							}
						} else {
							reader.close();
							throw new Exception("Invalid Target, PlannedTransition must target a \"State\", got : "
									+ target.getClass());
						}
					}
				}
			}
			reader.close();
		} catch (FileNotFoundException e) {
			// e.printStackTrace();
		} catch (IOException e) {
			// e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
			// if for some reasons, there are some errors, we don't initialize the driver
			// For example, if path == "", no file will be found
			// but it's because the user doesn't want to use a Driver so the path wasn't initialize
			return null;
		}

		return plannedUpdates;
	}

	// This function apply the PlannedUpdate under certain conditions
	public void apply(final double timeStep) {
		// precision : use to compare two Double variable
		double precision = 0.00001;
		if (timeStep == START_TIME
				|| (timeStep >= START_TIME && LOOP && (MAX_ITERATION == -1 || currentIteration < MAX_ITERATION)
						&& ((timeStep - START_TIME) % DELAY - 0) < precision)) {

			switch (ACTION) {
			case ADD:
				TARGET.add(lastValue);
				break;

			case SUBSTRACT:
				TARGET.substract(lastValue);
				break;

			case RESET:
				TARGET.reset();
				break;

			default:
				throw new RuntimeException(ACTION.toString() + " is an invalid action");
			}

			currentIteration++;
			lastValue *= COEFFICIENT;
		}
	}

}
