package goaml.mvc.view;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Random;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

public class DisplayView extends JFrame {

	private static final long serialVersionUID = 1L;

	private final Hashtable<String, HashSet<Dot>> DOTS;
	private final Hashtable<String, Boolean> RENDERABLE;
	private final Hashtable<String, Color> COLORS;

	// Variables which will be used for setting the position of the dots
	// according to the width and height of the ViewSimulation JPanel
	private int maxTime;
	private int maxValue;

	// Views (Inner class)
	private ChartPanel chartPanel;
	private LegendPanel legendPanel;

	public DisplayView(final String path) {
		this.DOTS = new Hashtable<String, HashSet<Dot>>();
		String line = "";
		String separator = ",";
		
		boolean header = true;

		BufferedReader reader = null;
		String[] names = null;
		int time = 0;
		try {
			reader = new BufferedReader(new FileReader(path));
			while ((line = reader.readLine()) != null) {
				if (header) {
					// Heading - Name of the state
					names = line.split(separator);
					for (int i = 0; i < names.length; i++) {
						this.DOTS.put(names[i], new HashSet<Dot>());
					}
					header = false;
				} else {
					// Body - Value of the state at different time step
					String[] values = line.split(separator);
					for (int i = 0; i < values.length; i++) {
						this.DOTS.get(names[i]).add(new Dot(time, Double.valueOf(values[i])));
					}
					time++;
				}
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		this.RENDERABLE = new Hashtable<String, Boolean>();
		this.COLORS = new Hashtable<String, Color>();

		this.maxTime = time - 1;
		this.maxValue = 0;

		// The seed is used in order to have the same set of int every time
		// "10" gives nice set of color
		Random random = new Random(10);
		for (String key : DOTS.keySet()) {
			// By default, every state is displayed
			RENDERABLE.put(key, true);
			// Choose a random color for each states
			// 0xFFFFFF is the maximum value of a color in RGB mode
			COLORS.put(key, new Color(random.nextInt(0xFFFFFF)));
		}

		// Views
		this.chartPanel = new ChartPanel();
		this.legendPanel = new LegendPanel();

		// Frame
		this.setTitle("Goaml");
		this.setLayout(new FlowLayout());
		this.add(chartPanel);
		this.add(legendPanel);
		this.pack();
		this.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		this.setLocationRelativeTo(null);
		this.setVisible(true);
	}

	private void render() {
		this.chartPanel.render();
	}

	private class ChartPanel extends JPanel {

		private static final long serialVersionUID = 1L;

		public static final int WIDTH = 1000;
		public static final int HEIGHT = (int) (WIDTH / 16.0 * 9.0);

		private final int DOT_SIZE = 4;
		private final int MARK_SIZE = 5;

		private final int NB_TICKS_XAXIS = 10;
		private final int NB_TICKS_YAXIS = 10;

		private final int MARGIN_LEFT = 70;
		private final int MARGIN_RIGHT = 20;
		private final int MARGIN_TOP = 40;
		private final int MARGIN_BOTTOM = 70;

		private final int X_AXIS_LENGTH = WIDTH - MARGIN_LEFT - MARGIN_RIGHT;
		private final int Y_AXIS_LENGTH = HEIGHT - MARGIN_TOP - MARGIN_BOTTOM;

		public ChartPanel() {
			Dimension size = new Dimension(WIDTH, HEIGHT);
			this.setPreferredSize(size);
			this.setMinimumSize(size);
			this.setMaximumSize(size);
			this.setLayout(new FlowLayout());
		}

		public void render() {
			this.repaint();
		}

		@Override
		protected void paintComponent(Graphics g) {
			g.setColor(Color.WHITE);
			g.fillRect(0, 0, this.getWidth(), this.getHeight());

			// reset maxValue according to the highest value in all visible sets
			maxValue = 0;
			for (String key : DOTS.keySet()) {
				if (RENDERABLE.get(key)) {
					HashSet<Dot> dots = DOTS.get(key);
					for (Dot dot : dots) {
						if (maxValue < dot.getValue()) {
							maxValue = (int) (dot.getValue());
						}
					}
				}
			}

			// draw dots only if they are in a visible set
			for (String key : DOTS.keySet()) {
				if (RENDERABLE.get(key)) {
					HashSet<Dot> dots = DOTS.get(key);
					for (Dot dot : dots) {
						double x, y;

						if (maxTime != 0) {
							x = MARGIN_LEFT + (double) dot.getTime() / maxTime * X_AXIS_LENGTH;
						} else {
							x = MARGIN_LEFT;
						}

						if (maxValue != 0) {
							y = HEIGHT - dot.getValue() / maxValue * Y_AXIS_LENGTH - MARGIN_BOTTOM;
						} else {
							y = HEIGHT - MARGIN_BOTTOM;
						}

						// Get the color of the state associated with this dot
						g.setColor(COLORS.get(key));
						g.fillOval((int) x - DOT_SIZE / 2, (int) y - DOT_SIZE / 2, DOT_SIZE, DOT_SIZE);
					}
				}
			}

			g.setColor(Color.BLACK);
			// draw x axis
			g.drawLine(MARGIN_LEFT, HEIGHT - MARGIN_BOTTOM, WIDTH - MARGIN_RIGHT, HEIGHT - MARGIN_BOTTOM);
			// draw y axis
			g.drawLine(MARGIN_LEFT, MARGIN_TOP, MARGIN_LEFT, HEIGHT - MARGIN_BOTTOM);

			// Use to know the width and height of the font used by "g"
			// in order to center texts marks
			FontMetrics fontMetrics = g.getFontMetrics();

			// Draw ticks marks
			// xAxis
			if (maxTime > 0) {
				for (int i = 0; i < NB_TICKS_XAXIS; i++) {
					double value = i * (double) (maxTime) / (NB_TICKS_XAXIS - 1);
					double x = MARGIN_LEFT + value / maxTime * X_AXIS_LENGTH;
					double y = HEIGHT - MARGIN_BOTTOM;
					// tick mark
					g.drawLine((int) x, (int) y, (int) x, (int) y + MARK_SIZE);
					// tick text
					String text = String.format("%.2f", value);
					int textWidth = fontMetrics.stringWidth(text);
					g.drawString(text, (int) x - textWidth / 2, (int) y + 20);
				}
			}
			// yAxis
			if (maxValue > 0) {
				for (int i = 0; i < NB_TICKS_YAXIS; i++) {
					double value = i * (double) (maxValue) / (NB_TICKS_YAXIS - 1);
					double x = 0;
					double y = HEIGHT - value / maxValue * Y_AXIS_LENGTH - MARGIN_BOTTOM;
					// tick mark
					g.drawLine((int) x + MARGIN_LEFT - MARK_SIZE, (int) y, (int) x + MARGIN_LEFT, (int) y);
					// tick text
					String text = String.format("%.2f", value);
					int textWidth = fontMetrics.stringWidth(text);
					int textHeight = fontMetrics.getHeight();
					g.drawString(text, (int) MARGIN_LEFT - textWidth - 10, (int) y + textHeight / 3);
				}
			}

			// chart legend
			g.setFont(new Font("default", Font.BOLD, 12));
			String text = "Time Step";
			int textWidth = fontMetrics.stringWidth(text);
			int textHeight = fontMetrics.getHeight();
			g.drawString(text, MARGIN_LEFT + X_AXIS_LENGTH / 2 - textWidth / 2,
					HEIGHT - MARGIN_BOTTOM / 2 + textHeight);
			g.drawString("Size of the population", 0, MARGIN_TOP / 2);

			g.dispose();
		}
	}

	private class LegendPanel extends JPanel {

		private static final long serialVersionUID = 1L;

		public static final int WIDTH = 200;
		public static final int HEIGHT = ChartPanel.HEIGHT;

		public LegendPanel() {
			Dimension size = new Dimension(WIDTH, HEIGHT);
			this.setPreferredSize(size);
			this.setMinimumSize(size);
			this.setMaximumSize(size);
			this.setLayout(new BorderLayout());
			
			JPanel panelLegend = new JPanel();
			panelLegend.setLayout(new BoxLayout(panelLegend, BoxLayout.PAGE_AXIS));

			// For each state
			for (String key : DOTS.keySet()) {

				// JBUTTON - Change color of the dots of the state
				JButton button = new JButton();
				button.setPreferredSize(new Dimension(15, 15));
				button.setBackground(COLORS.get(key));
				button.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent e) {
						Color color = JColorChooser.showDialog(null, "Color Picker", COLORS.get(key));
						if (color != null) {
							COLORS.put(key, color);
							button.setBackground(color);
						}

						render(); // Actualize
					}
				});

				// JCHECKBOX - Change visibility of the dots of the state (display them or not)
				JCheckBox checkBox = new JCheckBox(key);
				checkBox.setSelected(RENDERABLE.get(key));
				checkBox.addItemListener(new ItemListener() {
					@Override
					public void itemStateChanged(ItemEvent e) {
						if (RENDERABLE.get(key)) {
							RENDERABLE.put(key, false);
						} else {
							RENDERABLE.put(key, true);
						}

						render(); // Actualize
					}
				});

				// JPANEL - Put the JBUTTON and the JCHECKBOX in the same row
				JPanel panel = new JPanel();
				panel.setLayout(new FlowLayout(FlowLayout.LEFT));
				panel.add(button);
				panel.add(checkBox);

				panelLegend.add(panel);
			}
			
			this.add(new JScrollPane(panelLegend), BorderLayout.CENTER);
		}
	}

}
